/* davis-putnam.pro
 *
 * Authors: Chris Crabbe and Aaron Stern
 *
 * This file contains an implementation of the Davis-Putnam algorithm for
 * testing satisfiability of a CNF expression.
 */


/*****************************************************************************
*                                  DAVIS-PUTNAM                              *
*****************************************************************************/

/** 
 *  davis-putnam
 *
 *  MAIN
 *
 *  This is the main rule that is invoked to test a CNF expression.
 *
 */
davis-putnam([]).
davis-putnam(CNF) :- reduceClauses(CNF,B),
                     eliminateTautologies(B,C),
                     eliminateSingletonClauses(C,D),
                     eliminatePureLiterals(D,E),
                     splitRecursively(E).
                   
                   
/**
 *  reduceClauses
 *
 *  STEP 1
 *
 *  This rule removes from 'In' all duplicate clauses, duplicate literals
 *  within a clause, clauses contianing 'true', and 'fail' literals, and
 *  unifies the result with 'Out'.
 *
 */
reduceClauses(In,Out) :- maplist(list_to_set, In, S),
                         list_to_set(S,T),
                         remove_instances_of(true,T,U),
                         remove_from_clauses(fail,U,Out).


/**
 *  eliminateTautologies
 *
 *  STEP 2
 *
 *  Removes from 'In' all clauses which contain both a literal and 
 *  its negation, and unifies the result with 'Out'.
 *
 */
eliminateTautologies(In,Out) :- sublist(not_tautology, In, Out).


/**
 *  eliminateSingletonClauses
 *
 *  STEP 3
 *
 *  If 'In' has a singleton clause ( that is, {x} is a member of 'In' for some 
 *  literal x), then remove all clauses containing x, and remove ~x from 
 *  all remaining clauses, and unify the result with 'Out'.  This process
 *  is repeated iteratively until it can no longer be performed.
 * 
 */
eliminateSingletonClauses([],[]).
eliminateSingletonClauses(In, Out) :- not(get_first_singleton( In, _ )),
                                      Out = In.
eliminateSingletonClauses(In, Out) :- get_first_singleton( In, X ),
                                      remove_instances_of(X, In, B),
                                      negate(X, NotX),
                                      remove_from_clauses( NotX, B, Tmp ),
                                      (In=Tmp -> Out=Tmp ; eliminateSingletonClauses( Tmp, Out )).


/**
 *  eliminatePureLiterals  
 *
 *  STEP 4
 *
 *  If there is a literal x appearing in 'In' such that its negation appears 
 *  nowhere in 'In', then remove all clauses containing x and unify the result
 *  with 'Out'.  This process is repeated until it can no longer be performed.
 *
 */
eliminatePureLiterals([],[]).
eliminatePureLiterals(In,Out) :- flatten_one_level(In,Flat),
                                 list_to_set(Flat, Set),
                                 generate_delete_set(Set, Del),
                                 delete_clauses(Del, In, Tmp),
                                 (In=Tmp -> Out=Tmp ; eliminatePureLiterals(Tmp,Out)).


/**
 *  splitRecursively
 *
 *  STEP 5
 *
 *  Choose a literal x in '[H|T]' and split it into three sets:
 *      A: all clauses containing x
 *      B: all clauses containing ~x
 *      C: all clauses containing neither x nor ~x
 *  Remove x and ~x from all clauses, and create two new clause sets:
 *      A': A union C
 *      B': B union C
 *  If either A' or B' is satisfiable, then the rule is satisfiable.
 *
 */
splitRecursively([]).
splitRecursively([H|T]) :- nth0(0,H,E),negate(E,NotE),
                           triple_split(E,[H|T],A,B,C),
                           remove_from_clauses(E,A,A1),
                           remove_from_clauses(NotE,B,B1),
                           append(A1,C,L1),append(B1,C,L2),
                           (davis-putnam(L1) | davis-putnam(L2)).


/*****************************************************************************
*                              Helper Functions                              *
*****************************************************************************/

not_tautology(A) :- not(is_tautology(A)).
is_tautology([A|T]) :- (negate(A,B), member(B,T)) | is_tautology(T).

/**
 *  given a clause A, return [A]
 *  given [A], return A
 *
 *  return value stored in B
 */
negate([A|_],A).
negate(A,B) :- not(is_list(A)),
               B = [A|[]].


/**
 * remove all clauses in [H|T] which contain A and store the remains in B
 *
 */
remove_instances_of(_,[],[]).
remove_instances_of(A,[H|T], B) :- member(A,H),
                                   remove_instances_of(A, T, B).
remove_instances_of(A,[H|T], B) :- not(member(A,H)),
                                   remove_instances_of(A, T, C),
                                   B = [H|C].


/**
 *  remove E from any clause that contains it in [H|T] and store the remains
 *  in B
 *
 */
remove_from_clauses(_,[],[]).
remove_from_clauses(E,[H|T],B) :- delete(H,E,A),
                                  remove_from_clauses(E,T,C),
                                  B = [A|C].


/**
 * is A a singleton clause? 
 *
 */
singleton(A) :- is_list(A),
                length( A, 1 ).

/**
 *  Scan through the list and return the member of the first singleton clause
 *
 *  The first argument is the list, the second where the return value will be 
 *  stored
 *
 */
get_first_singleton([H|_], X) :- singleton(H),
                                 [X]=H.
get_first_singleton([H|T], X) :- not(singleton(H)),
                                 get_first_singleton( T, X ).



flatten_one_level([],[]).
flatten_one_level([H|T],F) :- flatten_one_level(T,A),
                              append(H,A,F).



delete_clauses([],A,A).
delete_clauses([E|T],L,O) :- remove_instances_of(E,L,A),
                             delete_clauses(T,A,O).

generate_delete_set([],[]).
generate_delete_set([E|T],S) :- negate(E,N), member(N,T),
                                select(N,T,A),
                                generate_delete_set(A,S).
generate_delete_set([E|T],S) :- negate(E,N), not(member(N,T)),
                                generate_delete_set(T,A),
                                S = [E|A].
				

/* triple_split
 *
 * Unification Arguments:
 *   0 - A variable
 *   1 - The master list of clauses
 *   2 - The list of clauses containing the variable
 *   3 - The list of clauses containing the negation of the variable
 *   4 - The list of clauses containing neither the variable not its negation
 */
triple_split(_,[],[],[],[]).
triple_split(S,[H|T],X,Y,Z) :- member(S,H),
                               triple_split(S,T,A,B,C),
                               X=[H|A],Y=B,Z=C.
triple_split(S,[H|T],X,Y,Z) :- negate(S,N),
                               member(N,H),
                               triple_split(S,T,A,B,C),
                               X=A,Y=[H|B],Z=C.
triple_split(S,[H|T],X,Y,Z) :- negate(S,N),
                               not(member(S,H)),not(member(N,H)),
                               triple_split(S,T,A,B,C),
                               X=A,Y=B,Z=[H|C].


